#include <stdio.h>
#include <netinet/in.h>		/* htonl */
#include "util.h"
#include "pool.h"
#include "type.h"
#undef USE_PIECE
#include "readbuf.h"
#include <stdlib.h>

static char *utf8(FILE *fp)
{
	int i;
	u2 length = read_u2(fp);
	char *buf = malloc(length+1);
	for ( i=0; i<length; i++ )
		buf[i]=getc(fp);
	buf[i]='\0';
	return buf;
}

static struct cp_info *read_pool( FILE *fp, int count )
{
	int i;
	for( i=1; i<=count; i++ ) {
		u1 tag = getc(fp);
		switch ( tag ) {
		case tag_UTF8: {
			char *str = utf8(fp);
			printf("%d: utf8: \"%s\"\n", i, str);
			free(str);
			break;
		}

		case tag_INT:
			printf("%d: int: %d\n", i, read_u4(fp));
			break;
			
		case tag_CLASS:
			printf("%d: class: %u(name_index)\n", i, read_u2(fp));
			break;

		case tag_STRING:
			printf("%d: string: %u(string_index)\n", i, read_u2(fp));
			break;

		case tag_FIELDREF:
			printf("%d: field ref: %d(class_index)", i, read_u2(fp));
			printf("%d(name_and_type_index)\n", read_u2(fp));
			break;

		case tag_METHODREF:
			printf("%d: method ref: %d(class_index)", i, read_u2(fp));
			printf("%d(name_and_type_index)\n", read_u2(fp));
			break;
			
		case tag_INTFCMETHREF:
			printf("%d: interface method ref: %d(class_index), ", i, read_u2(fp));
			printf("%d(name_and_type_index)\n", read_u2(fp));
			break;

		case tag_NAMEANDTYPE:
			printf("%d: name and type: %d(name_index), ",  i, read_u2(fp));
			printf("%d(descriptor_index)\n", read_u2(fp));
			break;

		default:
			fetal_error("unrecognize constant pool tag.");
			break;
		}
	}
	return NULL;
}


static void interface( FILE *fp, int count )
{
	int i;
	for ( i=0; i<count ; i++ ) {
		printf(" %d\n", read_u2(fp));
	}
}

static void attribute( FILE *fp )
{
	int length;
	int i;
	printf("  attribute_name_index: %d\n", read_u2(fp));
	length = read_u4(fp);
	printf("  attribute_length: %d\n", length);

	for ( i=0; i<length; i++ ) {
		printf(" %d", getc(fp));
	}
	putchar('\n');
}

void classdesc( FILE *fp )
{
	int count;
	struct cp_info *pools;
	int i;

	if ( read_u4(fp) != 0xCAFEBABE ) {
		fetal_error("not class file.");
	}

	printf("minor version: %d\n", read_u2(fp));
	printf("major version: %d\n", read_u2(fp));

	count = read_u2(fp);
	printf("constant_pool_count: %d\n", count);
	pools =	read_pool( fp, count-1 );

	printf("access_flags: %x\n", read_u2(fp));
	printf("this_class: %d\n", read_u2(fp));
	printf("super_class: %d\n", read_u2(fp));
	count = read_u2(fp);
	printf("interface_count: %d\n", count);
	interface(fp,count);

	count = read_u2(fp);

	printf("field_count: %d\n", count);
	for ( i=0; i<count; i++ ) {
		int attcnt;
		int j;
		printf(" access_flags: %x\n", read_u2(fp));
		printf(" name_index: %d\n", read_u2(fp));
		printf(" descriptor_index: %d\n", read_u2(fp));
		attcnt = read_u2(fp);
		printf(" atributes_count: %d\n", attcnt);
		for ( j=0; j<attcnt; j++ ) {
			attribute(fp);
		}
	}

	count = read_u2(fp);
	printf("methods_count: %d\n", count);

	for ( i=0; i<count; i++ ) {
		int attcnt;
		int j;
		printf(" access_flags: %x\n", read_u2(fp));
		printf(" name_index: %d\n", read_u2(fp));
		printf(" descriptor_index: %d\n", read_u2(fp));
		attcnt = read_u2(fp);
		printf(" atributes_count: %d\n", attcnt);
		for ( j=0; j<attcnt; j++ )
			attribute(fp);
	}

	count = read_u2(fp);
	printf("attributes_count: %d\n", count);

	for ( i=0; i<count; i++ )
		attribute(fp);
}

int main(int argc, char **argv)
{
	FILE *fp;
	if ( argc<2 ) {
		fprintf( stderr, "usage: %s classfile\n", argv[0] );
		return 1;
	}

	fp = fopen( argv[1], "r" );
	if ( fp == NULL ) {
		perror(argv[0]);
		return 1;
	}

	classdesc( fp );
	fclose(fp);
	return 0;
}
