#include "execute.h"
#include "type.h"
#include <piece.h>
#include "class.h"
#include "array.h"
#include "instance.h"
#include "jstring.h"
#include "error.h"

struct jclass *current_class;
struct method_header *current_method;

/*
 * abbreb.
 * nnt - name and type
 * cls - class 
 */

/*
 * $B%U%l!<%`(B
 *
 * ($B0z?t(B), ... , ($B%m!<%+%kJQ?t(B), ..., ($BA0$N(Bfp), (return address)
 * ^-fp                                         ^-sp($B=i4|CM(B)
 */

#define STACK_SIZE (4096*2)

u4 op_stack[STACK_SIZE];

u4 *fp;
const u1 *ip;
u4 *sp=op_stack;
jvalue retval;

#include "inst.h"
#include "execute_support.c"

void execute( )
{
	while (1) {
		switch ( *ip ) {

		case ACONST_NULL:
			PUSH(0);
			goto next1;
			
		case ALOAD_0:
		case ALOAD_1:
		case ALOAD_2:
		case ALOAD_3:
			aload( (*ip)-ALOAD_0 );
			goto next1;

		case ILOAD_0:
		case ILOAD_1:
		case ILOAD_2:
		case ILOAD_3:
			aload( (*ip)-ILOAD_0 );
			goto next1;

		case ALOAD:
		case ILOAD:
			aload( (*ip+1) );
			goto next2;

		case ASTORE_0:
		case ASTORE_1:
		case ASTORE_2:
		case ASTORE_3:
			astore( (*ip)-ASTORE_0 );
			goto next1;

		case ISTORE_0:
		case ISTORE_1:
		case ISTORE_2:
		case ISTORE_3:
			astore( (*ip)-ISTORE_0 );
			goto next1;

		case INEG:
			*sp = -*sp;
			goto next1;
			
		case ISTORE:
		case ASTORE:
			astore( *(ip+1) );
			goto next2;

		case AASTORE:
		case IASTORE: {
			jvalue val = pop();
			u4 idx = (u4)pop();
			struct array_head *h = (struct array_head *)pop();

			h->field[idx] = val;
			
			goto next1;
		}

		case BASTORE: {
			char byte = pop();
			u4 idx = (u4)pop();
			struct array_head *h = (struct array_head *)pop();
			char *val = (char*)h->field;

			val[idx] = byte;

			goto next1;
		}

		case AALOAD:
		case IALOAD: {
			u4 idx = (u4)pop();
			struct array_head *h = (struct array_head *)pop();

			PUSH( h->field[idx] );

			goto next1;
		}

		case BALOAD:
		{
			u4 idx = (u4)pop();
			struct array_head *array=(struct array_head*)pop();
			char *val = (char*)(array->field);
			
			PUSH( val[idx] );

			goto next1;
		}

		case ICONST_M1:
		case ICONST_0:
		case ICONST_1:
		case ICONST_2:
		case ICONST_3:
		case ICONST_4:
		case ICONST_5:
			PUSH( (*ip)-ICONST_0 );
			goto next1;
				

		case BIPUSH:
		{
			int i = *((signed char*)ip+1);
			PUSH(i);
			goto next2;
		}

		case SIPUSH:
		{
			int i = ((ip[1]<<8) | (ip[2]));
			PUSH(i);
			goto next3;
		}
			
		case PUTSTATIC: {
			jvalue *valp = static_field();
			*valp = pop();
			goto next3;
		}

		case GETSTATIC: {
			jvalue *valp = static_field();
			PUSH(*valp);
			goto next3;
		}

		case PUTFIELD: {
			jvalue v = pop();
			jvalue *valp = instance_field();
			*valp = v;
			goto next3;
		}

		case GETFIELD: {
			jvalue *valp = instance_field();
			PUSH(*valp);
			goto next3;
		}

		case IRETURN:
		{
			int i=pop();
			do_return();
			if ( ip==NULL )
				return;
			PUSH(i);
			goto next;
		}
			

		case RETURN:
		{
			do_return();
			/* main $B$+$i%j%?!<%s$9$k$H$-$N$?$a$K!$%9%?%C%/:G2<0L$K$O(BNULL$B$r@Q$s$G$*$/(B */
			if ( ip==NULL )
				return;
			goto next;
		}


#define IFJMP(v1,OP,v2) {\
	if ( ((int)v1) OP ((int)v2) )\
		ip+=(signed short)((*(ip+1)<<8) | (*(ip+2)));\
	else\
		goto next3;\
} goto next
			
		case IF_ICMPEQ:
			IFJMP(pop(), ==, pop());
		case IF_ICMPNE:
			IFJMP(pop(), !=, pop());
		case IF_ICMPLT: {
			int i2=pop();
			int i1=pop();
			IFJMP(i1, <, i2);
		}
		case IF_ICMPGE: {
			int i2=pop();
			int i1=pop();
			IFJMP( i1, >=, i2 );
		}
		case IF_ICMPGT: {
			int i2=pop();
			int i1=pop();
			IFJMP( i1, >, i2 );
		}
		case IF_ICMPLE: {
			int i2=pop();
			int i1=pop();
			IFJMP( i1, <=, i2 );
		}

		case IFEQ:
			IFJMP( pop(), ==, 0 );

		case IFNE:
			IFJMP( pop(), !=, 0 );

		case IFLT:
			IFJMP( pop(), <, 0 );

		case IFGE:
			IFJMP( pop(), >=, 0 );

		case IFGT:
			IFJMP( pop(), >, 0 );

		case IFLE:
			IFJMP( pop(), <=, 0 );

		case GOTO:
			ip += (signed short)((ip[1]<<8)|ip[2]);
			goto next;

		case INVOKESTATIC: {
			struct method_header *m = get_static_methref();
			ip+=3;
			invoke( m );
			goto next;
		}

		case INVOKESPECIAL:
		case INVOKEVIRTUAL: {
			invoke_instance();
			goto next;
		}

		case LDC:
		{
			int utf8idx = current_class->constants[ip[1]].val1;
			char* str = (char*)current_class->constants[utf8idx].val1;
			REF_PUSH(new_jstring( str ));

			goto next2;
		}

		case DUP:
			PUSH(*sp);
			goto next1;

		case IINC:
			fp[ip[1]]+=ip[2];
			goto next3;

		case IADD:
			PUSH(pop()+pop());
			goto next1;

		case ISUB: {
			int i2=pop();
			int i1=pop();
			PUSH( i1-i2 );
			goto next1;
		}

		case IMUL:
			PUSH( pop()*pop() );
			goto next1;

		case IAND:
			PUSH( pop()&pop() );
			goto next1;

		case IOR:
			PUSH( pop()|pop() );
			goto next1;

		case IXOR:
			PUSH( pop()^pop() );
			goto next1;

		case ANEWARRAY: {
			struct array_head *a = new_array(pop(),REF_ARRAY);
			REF_PUSH( a );
			goto next3;
		}
			
		case NEWARRAY: {
			int type;
			struct array_head *a;
			
			if ( ip[1] == T_BYTE )
				type = BYTE_ARRAY;
			else
				type = INT_ARRAY;

			a = new_array( pop(), type );
			REF_PUSH( a );
			goto next2;
		}

		case ARRAYLENGTH: {
			struct array_head *a = (struct array_head*)pop();
			UNREF( a );
			PUSH( a->length );
			goto next1;
		}			

		case POP:
			sp--;
			goto next1;
			

		case NEW: {
			int idx = (ip[1]<<8) |  ip[2];
			struct jclass *cls = (struct jclass*)(current_class->constants[idx].val1);
			struct jinstance *p = new_instance( cls );

			REF_PUSH( p );

			goto next3;
		}

		case ATHROW: {
			char errmes[256];
			struct jinstance *throwed = (struct jinstance *)pop();
			if ( unwind_exception(throwed->ofclass)<0 ) {
				sprintf( errmes, "uncahght exceptoin %s\n",
					 throwed->ofclass->classname );
				fetal_error( errmes );
			}

			PUSH(throwed);

			continue; /* ip$B$O@5$7$/$J$C$F$k$O$:(B */
		}
			
		default:
		{
			char errmes[256];
			sprintf( errmes, "unrecognized byte code: %d\n", *ip);
			fetal_error( errmes );
			return;
		}
		}

	next3:
		ip++;
	next2:
		ip++;
	next1:
		ip++;
	next:
		;
	}
		
}
