#include "pool.h"

static inline jvalue *static_field( )
{
	int idx = (*(ip+1)<<8) | *(ip+2);
	struct jclass *c=(struct jclass*)current_class->constants[idx].val1;
	int offset = current_class->constants[idx].val2;

	return &c->class_field_val[offset];
}

static inline jvalue *instance_field()
{
	int idx = (*(ip+1)<<8) | *(ip+2);
	int offset = current_class->constants[idx].val2;

	struct jinstance *i = (struct jinstance *)pop();
	return &i->field[offset];
}

static inline struct method_header *get_static_methref( )
{
	int idx = (*(ip+1)<<8) | *(ip+2);
	struct const_pool *mref = &current_class->constants[idx];
	struct jclass *cls;
	int nntidx, nameidx, typidx;
	struct method_header *m;
	const char *methname, *desc;

	if ( mref->type == tag_RESOLVED_METHOD )
		return (struct method_header *)mref->val1;

	cls = (struct jclass*)current_class->constants[mref->val1].val1;
	
	nntidx = mref->val2;
	nameidx = current_class->constants[nntidx].val1;
	typidx =current_class->constants[nntidx].val2;

	methname = STRING_CONST( current_class, nameidx );
	desc = STRING_CONST( current_class, typidx );

	m = get_static_method( cls, methname, desc );

	mref->type = tag_RESOLVED_METHOD;
	mref->val1 = (u4)m;

	return m;
}

/* $B%a%b(B.
 *
 * ($B0JA0$N%U%l!<%`(B), ... , ($B2?$+(B), ($B%U%l!<%`(B), ($BA0$N(Bfp), ($BA0$N%a%=%C%I(B), (return address)
 *                                 ^-fp                                  ^-sp
 * ($B0JA0$N%U%l!<%`(B), ... , ($B2?$+(B), ($B%U%l!<%`(B), ($BA0$N(Bfp), ($BA0$N%a%=%C%I(B), (return address)
 * ^-fp                    ^-sp
 */
static inline void do_return()
{
	u4 *tmpsp;
	ip = (u1*)pop();

	tmpsp = fp-1;

	current_method = (struct method_header *)pop();
	current_class = current_method->ofclass;

	fp = (jvalue*)pop();
	sp = tmpsp;
}

/* $B%a%b(B
 *
 * (local) , ... , ($B0z?t(B0), ... , ($B0z?t(Bn)
 * ^-old_fp                       ^-sp
 * (local) , ... , ($B0z?t(B0), ... , ($B0z?t(Bn), (local0), ..., (localn), (old_fp), (prev_method), (retaddr),
 *                 ^-fp                                                                      ^-sp
 */
static inline void invoke( struct method_header *m )
{
	if ( m->is_native ) {
		m->ncode();
		sp -= m->argnum;
		if ( m->is_retval )
			PUSH(retval);
	} else {
		jvalue *old_fp = fp;
		fp = sp-(m->argnum-1);
		sp = fp+m->max_local;
		
		PUSH(old_fp);
		PUSH(current_method);
		PUSH(ip);
		
		current_method = m;
		current_class = m->ofclass;
		
		ip = m->code;
	}
}

static inline void invoke_instance()
{
	struct const_pool *mref = &current_class->constants[(ip[1]<<8) | ip[2]];
	struct method_header *m;

	if ( (mref->type == tag_RESOLVED_METHOD) && (*ip == INVOKEVIRTUAL) ) {	/* $BL>A0$O2r7h:Q$_(B */
		int idx = (int)mref->val1;
		m = ((struct jinstance *)sp[-mref->val2])->ofclass->vtbl[idx];
	} else {
		struct jclass *c = (struct jclass *)current_class->constants[mref->val1].val1;
		
		struct const_pool *nnt = &current_class->constants[mref->val2];
		const char *mname = (const char *)current_class->constants[nnt->val1].val1;
		const char *desc = (const char *)current_class->constants[nnt->val2].val1;
		int argnum = compute_argnum(desc);
		
		if ( *ip == INVOKESPECIAL )
			m = get_method( c, mname, desc, GET_METHOD_SPECIAL );
		else
			m = get_method( ((struct jinstance *)sp[-argnum])->ofclass,
					mname, desc, GET_METHOD_VIRTUAL );

		if ( m==NULL ) {
			char errmes[256];
			
			sprintf( errmes, "not found method: class %s, method %s\n", c->classname, mname);
			fetal_error(errmes);
		}

		/* $BL>A0$N2r7h!%(B method_ref $B$N(B val1 $B$r(Bvtbl$B$N%*%U%;%C%H$K(B
		 * val2 $B$r0z?t$N?t$K(B
		 * type $B$r(B RESOLVED_METHOD $B$K(B
		 */

		if ( *ip == INVOKEVIRTUAL ) {
			mref->type = tag_RESOLVED_METHOD;
			mref->val1 = m->offset;
			mref->val2 = m->argnum-1; /* sp$B$+$i$N%*%U%;%C%H$@$+$i(B-1$B$7$F$*$/!%(B */
		}

	}

	ip+=3;
	invoke( m );
}
