#include "gc.h"
#include "usbdbg.h"
#include "alloc.h"
#include "error.h"

#include "execute.h"

/* pigc $B$h$j$b(B $B$3$C$A$N$[$&$,B.$$$h$&$J5$$,$9$k$1$I!"B,Dj$7$F$_$k$Y$-$+$b$7$l$J$$(B */

struct gcpro *protect_list = NULL;

#define OBJ_NUM (128)
#define OBJS_END (objs+OBJ_NUM)

static struct object_header objs[OBJ_NUM];
static struct object_header *top=objs;

static inline int is_obj_ref ( struct object_header *ptr ) {
	if ( ((u4)ptr)&0x03 )
		return 0;
	if ( (ptr>=objs) && (ptr<OBJS_END) )
		return 1;
	return 0;
}

void init_gc( )
{
	int i;

	for ( i=0; i<OBJ_NUM-1; i++ ) {
		objs[i].next = &(objs[i+1]);
		objs[i].alloced = 0;
	}

	objs[i].next = NULL;
	top = objs;
}

static void mark_object( struct object_header *ptr )
{
	int i;

	if ( is_obj_ref(ptr)==0 )
		return;

	if ( ptr->mark )
		return;

	ptr->mark = 1;

	if ( ptr->not_ref )
		return;

	for ( i=0; i<ptr->length; i++ ) {
		mark_object( (struct object_header*)ptr->field[i] );
	}
}

static inline void free_obj( struct object_header *obj )
{
	if ( obj->length!=0 )
		xfree( obj->field );
	obj->next = top;
	obj->alloced = 0;
	top = obj;
}

void gc()
{
	struct gcpro *pro = protect_list;
	u4 *ptr;
	struct object_header *objp;

	// fputs( "execute gc..\n", stderr );

	while ( pro ) {
		mark_object( pro->obj );
		pro = pro->next;
	}

	for ( ptr=op_stack; ptr<=sp; ptr++ ) {
		mark_object( (struct object_header *)(*ptr) );
	}

	for ( objp=objs; objp<OBJS_END; objp++ ) {
		if (objp->alloced==0) {
			continue;
		}
		if (objp->mark==1) {
			objp->mark = 0;
			continue;
		}
		
		free_obj( objp );
	}
}

struct object_header *alloc_object( int length )
{
	struct object_header *obj;

	if ( top == NULL ) {
		gc();

		if ( top == NULL )
			fetal_error("alloc failed.");
	}

	obj = top;
	top = top->next;

	obj->length = length;
	if ( length != 0 )
		obj->field = xcalloc( sizeof(jvalue), length+1 );

	obj->mark = 0;
	obj->alloced = 1;
	obj->ref_cnt = 1;

	return obj;
}
