#include "jstring.h"
#include "array.h"
#include <string.h>
#include "execute.h"
#include "instance.h"
#include "native.h"

#include "gc.h"

#define JSTRING_VTBL_SIZE (JOBJECT_VTBL_SIZE+2)

static struct jclass javalangString;

static void init_new()
{
	struct jinstance *str = (struct jinstance *)sp[0];
	struct array_head *val = (struct array_head*)new_array(1, BYTE_ARRAY);
	val->field[0]='\0';

	str->field[JSTRING_CONTENTS]=(jvalue)val;
	str->field[JSTRING_LENGTH]=0;
}

struct jinstance *new_jstring ( const char *str )
{
	int len = strlen(str);
	struct gcpro gcpro1;

	struct jinstance *newstr = new_instance(&javalangString);
	struct array_head *val;
	GCPRO1( newstr );

	val = new_array(len + 1,BYTE_ARRAY);
	strcpy( (char*)(val->field), str );

	newstr->field[JSTRING_CONTENTS] = (jvalue)val;
	newstr->field[JSTRING_LENGTH] = len;

	UNGCPRO;

	return newstr;
}

static void init_copy()
{
	struct jinstance *newstr = (struct jinstance *)sp[-1],
		*srcstr = (struct jinstance *)sp[0];
	const char *src = (char*)(((struct array_head*)(srcstr->field[JSTRING_CONTENTS]))->field);
	const int length = srcstr->field[JSTRING_LENGTH];

	struct array_head *val = new_array(length + 1, BYTE_ARRAY);

	strcpy( (char*)(val->field), src );
	newstr->field[JSTRING_CONTENTS] = (jvalue)val;
	newstr->field[JSTRING_LENGTH] = length;
}

static struct method_header methods[] = {
        NATIVE_METHOD("<init>","()V",1,0,init_new,&javalangString,JOBJECT_VTBL_SIZE),
	NATIVE_METHOD("<init>","(Ljava/lang/String;)V",2,0,init_copy, &javalangString,JOBJECT_VTBL_SIZE+1),
};

static struct method_header *vtbl[JSTRING_VTBL_SIZE];

static struct field_bucket field[2] = {
	{"contents",JSTRING_CONTENTS},
	{"length",JSTRING_LENGTH},
};

static struct jclass javalangString = NATIVE_CLASS_NOSTATIC( &javalangObject,"java/lang/String", field,
							     2, methods, vtbl );

struct jclass *new_jstring_class()
{
	int i;
	memcpy( vtbl, javalangObject.vtbl, 4*JOBJECT_VTBL_SIZE );
	for ( i=0; i<JSTRING_VTBL_SIZE-JOBJECT_VTBL_SIZE; i++ ) {
		vtbl[i+JOBJECT_VTBL_SIZE] = &methods[i];
	}
	return &javalangString;
}
