#include <stdlib.h>

#include "piece_bmp.h"
#include "array.h"
#include "execute.h"
#include "instance.h"
#include "native.h"
#include "gc.h"
#include "jstring.h"

#include <piece.h>
#include "readbuf.h"
#include "error.h"

#define PIECE_BMP_VTBL_SIZE (0)

static struct jclass PieceBmp;

enum {
	PIECEBMP_PIXELS,
	PIECEBMP_WID_HT,
};

static inline int read_bmp( const char *fname, struct array_head **pixels,
			    int *width, int *height )
{
	PBMP_FILEHEADER bmph;
	READ_BUF *buf = readbuf_open( fname );
	int w,h;

	if ( buf == NULL )
		return -1;

	read_buf( buf, sizeof(bmph), &bmph );

	*width = bmph.w;
	*height = bmph.h;

	w = bmph.w;
	h = bmph.h;

	*pixels = new_array( (w*h/4) + (w*h/8) , BYTE_ARRAY );
			   /* pixel     mask */
	read_buf( buf, (w*h/4)+(w*h/8), (*pixels)->field );

	readbuf_close( buf );

	return 0;
}

/* PieceBmp(Ljava/lang/String) */
static void init()
{
	struct jinstance *bmp = INSTANCE(sp[-1]);
	struct jinstance *str = INSTANCE(sp[0]);

	const char *fname = JSTRING_GETCONTENT(str);
	int wid,ht;
	struct array_head *pixels;

	if ( read_bmp( fname, &pixels, &wid, &ht ) == -1 )
		fetal_error("file open error");

	bmp->field[ PIECEBMP_PIXELS ] = (jvalue)pixels;
	bmp->field[ PIECEBMP_WID_HT ] = (jvalue)((wid<<16)|(ht));
}

static struct method_header methods[] = {
        NATIVE_METHOD("<init>","(Ljava/lang/String;)V",2,0,
		      init,&PieceBmp,0),
};


static struct jclass PieceBmp = NATIVE_CLASS_NOSTATIC( &javalangObject, "vaca/PieceBmp", NULL,
						       2, methods, NULL );

struct jclass *init_piece_bmp_class()
{
	return &PieceBmp;
}

void piece_bmp_to_PIECE_BMP( struct jinstance *this_, PIECE_BMP *bmp )
{
	const u4 wh = this_->field[PIECEBMP_WID_HT];
	const u2 w = wh>>16;
	const u2 h = wh&0xffff;
	char *pixels = (char*)(ARRAY_VAL(this_->field[PIECEBMP_PIXELS]));

	*bmp = ((PIECE_BMP) {{0,0,2,1, w, h, 0},
			     pixels,
			     pixels+w*h/4} );
}
