/****
  pigc.c
  $Id: pigc.c,v 1.6 2003/02/28 08:42:42 ko1 Exp $

  Create : K.S. 03/02/14 15:41:52
*/

#include <piece.h>
#include <string.h>

#include "pigc.h"
#include "usbdbg.h"

#define MARK_TABLE_SIZE_PARAM  1

#if MARK_TABLE_SIZE_PARAM   == 4
#define MARK_SET_SHIFT         2
#elif MARK_TABLE_SIZE_PARAM == 1
#define MARK_SET_SHIFT         0
#else
error
#endif


#ifdef PIGC_DEBUG
PIGC_DEBUGINFO Pigc_DebugInfo;
#endif

void *global_HeapTop;
void *global_HeapEnd;


static void mark();
static int  sweep();
void pigc_heapdump();

#define HMO_GC_USER (HMO_DEFAULT_USER + 1)

/////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////
// heap func from P/ECE kernel

static void heapbroken( HEAPMEM *phm )
{
  pigc_heapdump();
  pceAppReqExit(0);
  // APIErr( (unsigned long)phm, PCEKN_ERR_HEAP_BROKEN );
}


static void SetHEAPMEM( HEAPMEM *phm, int owner, void *chain )
{
	phm->mark  = HEAPMEMMK;
	phm->owner = owner;
	phm->chain = chain;
}

static void CombineFreeHEAPMEM( HEAPMEM *phm )
{
	HEAPMEM *phm0 = phm;
	HEAPMEM *tphm;
  
	phm = phm->chain;
	while ( phm ) {
		if ( phm->mark != HEAPMEMMK ) heapbroken( phm );
		if ( phm->owner ) break;
    
    tphm = phm;
		phm = phm->chain;

    tphm->mark  = 0;
    tphm->chain = 0;
    // DbgOut("hoe");
		phm0->chain = phm;
	}
}

static void AllocHEAPMEM(HEAPMEM *phm1, unsigned long size, unsigned long size1){
	if(size1 >= size+sizeof(HEAPMEM)){
		HEAPMEM *phm2 = phm1->chain;
		SetHEAPMEM( phm1, HMO_GC_USER, ((unsigned char *)phm1)+size );
		SetHEAPMEM( phm1->chain, 0, phm2 );
		CombineFreeHEAPMEM( phm1->chain );
	}
	else{
		phm1->owner = HMO_GC_USER;
	}
}

void *pceHeapAlloc_u(unsigned long size){
	HEAPMEM *phm = global_HeapTop;
	HEAPMEM *phm1 = NULL;
	unsigned long ssize = 0xffffffff;
  unsigned long sz;
  
  // size : KvȃTCY + sizeof(HEAPMEM) AS̔{ɐ񂵂
	size = ((size + sizeof(HEAPMEM) + 3) & ~3);	// 4̔{ɐ

	while(phm){
    if(phm->mark != HEAPMEMMK){
      heapbroken(phm);
    }
    if(!phm->owner){
      sz = (unsigned long)phm->chain - (unsigned long)phm;
      if (sz >= size && size < ssize){
        ssize = sz;
        phm1 = phm;
      }
    }
    phm = phm->chain;
  }
  
	if(phm1){
		AllocHEAPMEM( phm1, size, ssize );
		phm1 = (HEAPMEM *)(((unsigned long)phm1)+sizeof(HEAPMEM));
	}
	return phm1;
}

/////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////

void *pigc_malloc(unsigned long size){
  void *mem;

#ifdef PIGC_DEBUG
  Pigc_DebugInfo.occur = 0;
#endif

  mem = pceHeapAlloc_u(size);
  
  if(mem == 0){
#ifdef PIGC_DEBUG
    unsigned long start  = pceTimerGetPrecisionCount();
    memset(&Pigc_DebugInfo,0,sizeof(Pigc_DebugInfo));
    Pigc_DebugInfo.occur = 1;
#endif
    
    mark();
    if(sweep() == 0){
      return 0;
    }

    mem = pceHeapAlloc_u(size);

#ifdef PIGC_DEBUG
    Pigc_DebugInfo.gc_time = pceTimerAdjustPrecisionCount(start,pceTimerGetPrecisionCount());

    // if(pigc_check() == 0){ Pigc_DebugInfo.errno = 15; return 0;}
    Pigc_DebugInfo.free    = pceHeapGetMaxFreeSize();

    // dbg \
    DbgOut("|free  : %d",pceHeapGetMaxFreeSize());
    DbgOut("|n1,n2 : %d , %d",Pigc_DebugInfo.heapn,Pigc_DebugInfo.freed);
    DbgOut("|item  : %d",Pigc_DebugInfo.heapn - Pigc_DebugInfo.freed);
    DbgOut("|mk,ck : %d , %d",Pigc_DebugInfo.mark ,Pigc_DebugInfo.check);
    DbgOut("|s,d,h : %d , %d , %d",Pigc_DebugInfo.m_stack,Pigc_DebugInfo.m_data,Pigc_DebugInfo.m_heap);
    DbgOut("|time  : %u",Pigc_DebugInfo.gc_time);
    DbgOut("|> %s",Pigc_DebugInfo.msg);
    // pigc_heapdump();
#endif
  }
  return mem;
}

int pigc_nongcmalloc(unsigned long size){
  HEAPMEM *mem = pigc_malloc(size);
  if(mem){
    (mem-1)->owner = HMO_DEFAULT_USER;
  }
  return 0;
}


int pigc_free(void *mem){
  return pceHeapFree(mem);
}


int pigc_init(){
  void *mem = pceHeapAlloc(1);
  if(mem){
    HEAPMEM *hm = (HEAPMEM *)((unsigned long)mem - sizeof(HEAPMEM));
    global_HeapTop = hm;
    while(hm && hm->owner != HMO_SYSTEM){
      hm = hm->chain;
    }
    global_HeapEnd = hm;

    pceHeapFree(mem);
    
    // q[v̏
    memset(mem,0,global_HeapEnd - global_HeapTop - sizeof(HEAPMEM));
    return 0;
  }
  else{
    DbgOut("mem init err");
    return -1;
  }
}


/////////////////////////////////////////////

unsigned int global_MarkBits[1024 * 2 / MARK_TABLE_SIZE_PARAM]; // 8KB / 4


#define HEAP_START 0x100000
#define HEAP_END   0x140000

// #define IS_POINTER(p) ((!((p) & 0x3)) && ((p) > (unsigned long)global_HeapTop) && ((p) < HEAP_END))
// ̂ق
#define IS_POINTER(p) (((p) > (unsigned long)global_HeapTop) && \
                       ((p) < (unsigned long)global_HeapEnd) && (!((p) & 0x3)))

#define STACK_START 0x1000
#define STACK_END   0x2000
#define DATA_START  (&__START_DEFAULT_DATA)

void MARK_SET(unsigned long p){
  unsigned long i,s;
  //i = (p - HEAP_START)  / (4 * 32);
  //s = ((p - HEAP_START) / 4) % 32;

  
  p -= HEAP_START;
  i  = p >> (7+MARK_SET_SHIFT);
  s  = (p >> (2+MARK_SET_SHIFT)) & 0x1f;
  
  // for debug


  global_MarkBits[i] |= 0x01 << s;

#ifdef PIGC_DEBUG
  if(i > sizeof(global_MarkBits) / sizeof(int)){
    Pigc_DebugInfo.errno = 3;
    pcesprintf(Pigc_DebugInfo.msg,"mk err \n%08x %d,%d\n%08x",p,i,s,global_HeapTop);
  }
  
  Pigc_DebugInfo.mark++;
#endif
}

int IS_MARKED(unsigned long p){
  unsigned long i,s;
  p -= HEAP_START;
  i  = p >> (7+MARK_SET_SHIFT);
  s  = (p >> (2+MARK_SET_SHIFT)) & 0x1f;

#ifdef PIGC_DEBUG
  Pigc_DebugInfo.check++;
#endif
  
  return global_MarkBits[i] & (0x01 << s);
}

typedef unsigned long * PTR;

void mark_heap(HEAPMEM *hm){
  PTR p = ((PTR)((unsigned long)hm + sizeof(HEAPMEM)));
  for(;(unsigned long)p < (unsigned long)hm->chain ;p++){
    if(IS_POINTER(*p)){
      if(*p < (unsigned long)hm && !IS_MARKED(*p)){
        // OQ  ܂}[NĂȂꍇ
        // ̃IuWFNgS蒼ȂƂȂ
        HEAPMEM *fhm = (HEAPMEM*)(((unsigned long)*p) - sizeof(HEAPMEM));

        if(fhm->mark  == HEAPMEMMK &&     // q[v[
           fhm->owner == HMO_GC_USER      // GC[U[
           ){
          // DbgOut("prev ref : %08x",fhm);
          mark_heap(fhm);
        }
      }
      // DbgOut("markH at [%08x - %08x] / %08x -> %08x",hm,hm->chain,p,*p);
      MARK_SET(*p);
#ifdef PIGC_DEBUG
      Pigc_DebugInfo.m_heap++;
#endif
    }
  }
}

//
extern unsigned long __START_DEFAULT_DATA;

static void mark(){
  PTR p;
  
  memset(global_MarkBits,0,sizeof(global_MarkBits));
  
  // stack mark
  for(p = (PTR)&p ;(unsigned long)p < STACK_END ; p++){
    if(IS_POINTER(*p)){
      // DbgOut("markS at %08x -> %08x",p,*p);
      MARK_SET(*p);
#ifdef PIGC_DEBUG
      Pigc_DebugInfo.m_stack++;
#endif
    }
  }
  // data mark
  for(p = (PTR)DATA_START ;p < (PTR)global_HeapTop ; p++){
    if(IS_POINTER(*p)){
      // DbgOut("markD at %08x -> %08x",p,*p);
      MARK_SET(*p);
#ifdef PIGC_DEBUG
      Pigc_DebugInfo.m_data++;
#endif
    }
  }

  // heap mark
  {
    HEAPMEM *hm = (HEAPMEM*)global_HeapTop;
    while(hm){
      if(hm->owner == HMO_GC_USER &&
         IS_MARKED((unsigned long)hm + sizeof(HEAPMEM))
         ){
        // DbgOut("marked H %08x",hm);
        mark_heap(hm);
      }
      hm = hm->chain;
    }
  }
}

static int sweep(){
  HEAPMEM *hm = (HEAPMEM*)global_HeapTop;

#ifdef PIGC_DEBUG
  int n1=0,n2=0;
#endif
  
  while(hm){
    
    if(hm->mark != HEAPMEMMK){
      DbgOut("mark error");

#ifdef PIGC_DEBUG
      pcesprintf(Pigc_DebugInfo.msg,"broken\n%d,%d<",n1,n2);
      Pigc_DebugInfo.errno = 1;
#endif
      return 0;
    }
    
    if(hm->owner == HMO_GC_USER){

#ifdef PIGC_DEBUG
      n1++;
#endif
      
      // check
      if(!IS_MARKED((unsigned long)hm + sizeof(HEAPMEM))){
        // free
        hm->owner = 0;

#ifdef PIGC_DEBUG
        n2++;
#endif
      }
    }
//    hmb = hm;
    hm  = hm->chain;
  }

  // combine
  hm = (HEAPMEM*)global_HeapTop;
  while(hm){
    if(hm->owner == 0){
      CombineFreeHEAPMEM(hm);
    }
    hm = hm->chain;
  }
  

#ifdef PIGC_DEBUG
  Pigc_DebugInfo.heapn = n1;
  Pigc_DebugInfo.freed = n2;
#endif

  return 1;
}

// `FbN
int pigc_check(){
  HEAPMEM *hm = (HEAPMEM*)global_HeapTop;
  int err = 0;
  while(hm){
    if(hm->mark != HEAPMEMMK){
      err = 1;
      break;
    }
    hm = hm->chain;
  }
  if(err){
    DbgOut("memory check error");
    pigc_heapdump();
#ifdef PIGC_DEBUG
    Pigc_DebugInfo.errno = 1;
#endif
  }
  return !err;
}

void pigc_heapdump(){
  HEAPMEM *hm = (HEAPMEM*)global_HeapTop;
  // So
  hm = (HEAPMEM*)global_HeapTop;
  DbgOut("heap dump");
  while(hm){
    DbgOut("heap %08x : %04x , %04x , %08x",hm,hm->mark,hm->owner,hm->chain);
    hm = hm->chain;
  }
}

int pigc_is_available(void *mem){
  HEAPMEM *hm = (HEAPMEM*)global_HeapTop;
  // So
  hm = (HEAPMEM*)global_HeapTop;
  while(hm){
    if((hm->owner & 3) && (unsigned long)mem == (unsigned long)hm+sizeof(HEAPMEM)){
      return 1;
    }
    hm = hm->chain;
  }
  return 0;
}

/**
  flag ̒l
  0 : m
  1 : GCȂݒ
  2 : GCݒ

  ԂlF
  0 : ȃ|C^Aflag  0,1,2 ȊO
  1 : O̐ݒ GCȂA
  2 : O̐ݒ GCA
 */
int pigc_setflag(void *mem,int flag){
  if(pigc_is_available(mem)){
    HEAPMEM *hm = (HEAPMEM*)((unsigned long)mem - sizeof(HEAPMEM));
    int prev;
    switch(flag){
    case 0:
      return hm->owner;
    case 1:
    case 2:
      prev = hm->owner;
      hm->owner = flag;
      return prev;
    }
  }
  return 0;
}