#include "readbuf.h"
#include "alloc.h"
#include <piece.h>
#include <string.h>

static void read_buffer( READ_BUF *buf )
{
	if ( buf->iszip ) {
		buf->bufp = buf->buf = buf->unziped+ buf->sect*4096 ;
	} else {
		pceFileReadSct( &buf->fp, NULL, buf->sect, 4096 );
		buf->buf = buf->fp.aptr;
		buf->bufp = buf->buf;
	}
}

u1 read_u1( READ_BUF *buf )
{
	if ( buf->bufp >= buf->buf+4096 ) {
		buf->sect++;
		read_buffer( buf );
	}

	return *(buf->bufp++);
}

u2 read_u2( READ_BUF *rbuf )
{
	char buf[2];
	int i;

	for ( i=0; i<2; i++ )
		buf[1-i] = read_u1( rbuf );

	return *(u4*)buf;
}

u4 read_u4( READ_BUF *rbuf )
{
	char buf[4];
	int i;

	for ( i=0; i<4; i++ )
		buf[3-i] = read_u1( rbuf );

	return *(u4*)buf;
}

size_t buf_tell( READ_BUF *buf )
{
	size_t ret = buf->sect*4096;
	ret += buf->bufp - buf->buf;

	return ret;
}

void buf_setpos( READ_BUF *buf, size_t offset )
{
	buf->sect = offset/4096;
	read_buffer( buf );

	buf->bufp = offset%4096+buf->buf;
}

void buf_seek( READ_BUF *buf, size_t offset )
{
	int newoffset = buf->sect*4096+buf->bufp-buf->buf + offset;
	buf_setpos( buf, newoffset );
}

void read_buf( READ_BUF *buf, size_t size, void *read_buf )
{
	int rest = buf->buf+4096 - buf->bufp;

	while ( 1 ) {
		if ( size <= rest ) {
			memcpy( read_buf, buf->bufp, size );
			buf->bufp += size;
			break;
		}

		buf->sect++;
		memcpy( read_buf, buf->bufp, rest );

		size -= rest;
		read_buffer( buf );
		rest = 4096;
	}
}

#include "zip.h"

READ_BUF *readbuf_open( const char *fname )
{
	READ_BUF *buf = xmalloc( sizeof(READ_BUF) );

	if ( pceFileOpen( &buf->fp, fname, FOMD_RD )!=0 ) {
		/* jar $B$+$i(B $B3+$$$F$_$k(B */
		if ( default_jar.readbuf != NULL ) {
			int ret = zip_readbuf_open( &default_jar, fname, buf );
			if ( ret == 0 )
				return buf;
		}

		xfree( buf ); 
		return NULL;
	}

	buf->sect=0;
	read_buffer( buf );
	buf->bufp = buf->buf;

	return buf;
}

#include "alloc.h"

void readbuf_close( READ_BUF *buf )
{
	if ( buf->iszip ) {
		xfree( buf->unziped );
	} else {
		pceFileClose( &buf->fp );
		xfree( buf );
	}
}
