#include "stringbuffer.h"
#include "jstring.h"
#include "instance.h"
#include "array.h"
#include "execute.h"
#include "native.h"
#include <string.h>

static struct jclass stringbuffer;

#define JSTRINGBUFFER_METHOD_CNT 7
#define JSTRINGBUFFER_VTBL_SIZE (JOBJECT_VTBL_SIZE+JSTRINGBUFFER_METHOD_CNT)

#define DEFAULT_LENGTH 16

#define CONTENTS(strbuf) ((struct array_head*)((strbuf)->field[STRINGBUFFER_CONTENTS]))
#define LENGTH(strbuf) ((strbuf)->field[STRINGBUFFER_LENGTH])
#define BUFLEN(strbuf) ((strbuf)->field[STRINGBUFFER_BUFLEN])

#define STRING_CONTENTS(str) ((char*)(((struct array_head*)((str)->field[JSTRING_CONTENTS]))->field))

static void init_buffer( struct jinstance *buffer, int len )
{
	struct array_head *buf;
	BUFLEN(buffer) = len;

	buf = new_array( len+1, BYTE_ARRAY );
	CONTENTS(buffer) = (jvalue)buf;
}

/* <init>(LJava/lang/String;) */
static void init_string_s()
{
	struct jinstance *string = (struct jinstance *)sp[0];
	struct jinstance *strbuf = (struct jinstance *)sp[-1];
	const char *val = STRING_CONTENTS(string);
	const int len = (const int)string->field[JSTRING_LENGTH];

	init_buffer( strbuf, len*2 );

	LENGTH( strbuf ) = len;
	strcpy( (char*)(CONTENTS(strbuf)->field), val );
} 

static void init_string()
{
	struct jinstance *this_ = (struct jinstance *)sp[0];
	init_buffer( this_, DEFAULT_LENGTH );
	LENGTH(this_) = 0;
	CONTENTS(this_)->field[0] = '\0';
}

static void append_int();

static void init_string_i()
{
	struct jinstance *this_ = (struct jinstance *)sp[-1];

	init_buffer( this_, DEFAULT_LENGTH );
	CONTENTS(this_)->field[0] = '\0';
	LENGTH(this_) = 0;

	append_int();
}

static void expand( struct jinstance *buf, size_t len )
{
	struct array_head *newbuf = new_array( len*2+1, BYTE_ARRAY );

	strcpy( (char*)(newbuf->field), (char*)(CONTENTS(buf)->field) );
	CONTENTS(buf) = newbuf;
	BUFLEN(buf) = len*2;
}

static void append(struct jinstance *buf, const char *src, size_t srclen)
{
	int newlen = LENGTH(buf)+srclen;
	char *p;

	if ( newlen > BUFLEN(buf) ) {
		expand( buf, newlen );
	}

	LENGTH(buf) = newlen;
	p = (char*)(CONTENTS(buf)->field);
	strncat( p, src, srclen );
}

/* append(Ljava/lang/String;) */
static void append_str()
{
	struct jinstance *this_ = (struct jinstance *)sp[-1];
	struct jinstance *str = (struct jinstance *)sp[0];

	append( this_,  STRING_CONTENTS(str),
		str->field[JSTRING_LENGTH] );

	retval=(jvalue)this_;
}

/* String toString() */
static void tostr()
{
	struct jinstance *this_ = (struct jinstance *)sp[0];
	struct jinstance *newstr = new_jstring( (char*)(CONTENTS(this_)->field) );

	retval = (jvalue)newstr;
}

/* UTF$B$@$+$i0c$&$h$J$!!D(B $B=$@5$7$?$$!%(B */
static void append_char()
{
	char buf[3];
	struct jinstance *this_ = (struct jinstance *)sp[-1];
	((u2*)buf)[0] = (short)sp[0];

	buf[2] = '\0';
	append ( this_, buf, 2 );

	retval = (jvalue)this_;
}

/* StringBuffer append(int i) */
static void append_int()
{
	char buf[16];		/* 12$B$G==J,$J$s$@$1$I!$%Q%i%N%$%"$i$7$$$G$9!%(B */
	struct jinstance *this_ = (struct jinstance *)sp[-1];
	int i = (int)sp[0];

	pcesprintf( buf, "%ld", i );
	append ( this_, buf, strlen(buf) );

	retval = (jvalue)this_;
}
					
static struct method_header methods[] = {
	NATIVE_METHOD("<init>","(Ljava/lang/String;)V",2,0,
		      init_string_s,&stringbuffer,JOBJECT_VTBL_SIZE),
	
	NATIVE_METHOD("append","(Ljava/lang/String;)Ljava/lang/StringBuffer;",2,1,
		      append_str,&stringbuffer,JOBJECT_VTBL_SIZE+1),

	NATIVE_METHOD("toString","()Ljava/lang/String;",1,1,
		      tostr,&stringbuffer,JOBJECT_VTBL_SIZE+2),

	NATIVE_METHOD("append","(C)Ljava/lang/StringBuffer;",2,1,
		      append_char,&stringbuffer,JOBJECT_VTBL_SIZE+3),

	NATIVE_METHOD("append","(I)Ljava/lang/StringBuffer;",2,1,
		      append_int,&stringbuffer,JOBJECT_VTBL_SIZE+4),

	NATIVE_METHOD("<init>", "(I)V", 2, 0,
		      init_string_i,&stringbuffer, JOBJECT_VTBL_SIZE+5),
	
	NATIVE_METHOD("<init>", "()V", 1, 0,
		      init_string,&stringbuffer, JOBJECT_VTBL_SIZE+6),
};

static struct field_bucket field[3] = {
	{"contents",STRINGBUFFER_CONTENTS},
	{"length",STRINGBUFFER_LENGTH},
	{"buflen",STRINGBUFFER_BUFLEN},
};

static struct method_header *vtbl[JSTRINGBUFFER_VTBL_SIZE];

static struct jclass stringbuffer = NATIVE_CLASS_NOSTATIC( NULL, "java/lang/StringBuffer",
							   field, 3, methods, vtbl );

struct jclass *new_stringbuffer_class()
{
	init_native_vtbl( vtbl, javalangObject.vtbl, JOBJECT_VTBL_SIZE,
		   methods, JSTRINGBUFFER_METHOD_CNT );

	return &stringbuffer;
}
