#include "zip.h"
#include "type.h"
#include "alloc.h"

#include "inflate.h"

#include <string.h>

#define ZIP_SIG 0x04034b50

struct zipheader {
	u4 sig;
	u2 needver;
	u2 flag;
	u2 method;
	u2 modtime;
	u2 moddata;		/* $B$3$3$^$G(B 14 */

	u2 reserved;		/* $BB?J,%Q%G%#%s%0(B */

	u4 crc32;
	u4 compsize;
	u4 uncompsize;
	u2 fnamelen;
	u2 exfieldlen;		/* $B$3$3$^$G(B 16 */
};

int zip_open( struct zipfile *zip, char *fname )
{
	if ( (zip->readbuf = readbuf_open( fname )) == NULL )
		return -1;

	return 0;	
}

#define MAX_FNAMELEN 128

int zip_readbuf_open( struct zipfile *zip, char *fname, READ_BUF *ret )
{
	struct zipheader head;
	char fnamebuf[MAX_FNAMELEN+1];

	char isfind = 0;
	
	buf_setpos( zip->readbuf, 0 );

	while ( 1 ) {
		
		read_buf( zip->readbuf, 14, &head );

		if ( head.sig != ZIP_SIG )
			return -1;

		read_buf( zip->readbuf, 16, &head.crc32 );

		if ( head.fnamelen>MAX_FNAMELEN )
			return -1;
		read_buf( zip->readbuf, head.fnamelen, fnamebuf );
		fnamebuf[head.fnamelen] = '\0';

		if ( strcmp( fnamebuf, fname ) == 0 )
			isfind = 1;

		buf_seek( zip->readbuf, head.exfieldlen );

		if ( isfind ) {
			ret->iszip = 1;
			ret->unziped = xmalloc( head.uncompsize );

			if ( inflate( zip->readbuf, ret->unziped )!= 0 ) {
				xfree( ret->unziped );
				return -1;
			}

			ret->bufp = ret->buf = ret->unziped;
			ret->sect = 0;

			return 0;
		} else {
			buf_seek( zip->readbuf, head.compsize );
		}
	}

	return -1;		/* $B$3$3$^$GMh$J$$(B */
}
	
